#include <cmath>
#include "resummation.hh"
#include "chebyshev.hh"
#include "QCD_EW.hh"
#include "math/special_functions.hh"
#include "math/integration.hh"
#include "math/struct_definitions.hh"
#include <gsl/gsl_sf.h>
#include <gsl/gsl_errno.h>
#include "math/cuba.h"

#include "../../ggHiggs/src/NLO_finite-mt/bdv.hh"
#include "../../ggHiggs/src/NNLO_finite-mt/nnlo_finite-mt.hh"

#include "LHAPDF/LHAPDF.h"


using namespace std;




namespace resumm {

  string sPrescr(prescription p) {
    string s;
    if(p == BP00) s = "BP00";
    if(p == BP0) s = "BP0";
    if(p == BP1) s = "BP1";
    if(p == BP2) s = "BP2";
    if(p == MP00) s = "MP00";
    if(p == MP0) s = "MP0";
    if(p == MP2) s = "MP2";
    return s;
  }
  string sAP(APtype ap) {
    string s;
    if(ap == AP0) s = "AP0";
    if(ap == AP1) s = "AP1";
    if(ap == AP2) s = "AP2";
    if(ap == AP12) s = "AP12";
    if(ap == AP ) s = "AP";
    return s;
  }


  ////////////////////////////////
  //        PUBLIC  PART        //
  ////////////////////////////////

  Resummation::Resummation(process p, collider c, string PDFset, int PDFsubset, verbosity v) {
    proc = p;
    coll = c;
#ifndef LHAPDF_MAJOR_VERSION
    LHAPDF::setVerbosity((LHAPDF::Verbosity)v);
#endif
    LHAPDF::initPDFSet(PDFset, PDFsubset);
    mtop = 173.;
    mbot = -1;
    mcha = -1;
    mtEFT = true;
    //
    // default setup: BorelPrescription resumming terms z * [log^k{(1-z)/sqrt(z)}/(1-z)]_+
    // and delta(1-z) terms as much as possible  (easiest most physical choice)
    prescr = BP2;
    APimpr = AP1;
    g0exp = true;
    expDconst = true;
    expBP = false;
    //shift = false;
    //
    // other init...
    init_mass_thresholds(LHAPDF::getQMass(4), LHAPDF::getQMass(5), LHAPDF::getQMass(6));
    SetBPpars();
    SetMPpars();
    //
    last_muFrat = __builtin_nan("");
    last_muRrat = __builtin_nan("");
    last_M      = __builtin_nan("");
    last_sqrts  = __builtin_nan("");
    last_nf     = __builtin_nan("");
    last_mtop   = __builtin_nan("");
    last_EFT    = __builtin_nan("");
    last_Ncheb  = __builtin_nan("");
    firstinit = true;
    //
    return;
  }
  Resummation::~Resummation() {
    //delete[] coeff;
    return;
  }

  void Resummation::PDFinit(string PDFset, int PDFsubset, verbosity v) {
#ifndef LHAPDF_MAJOR_VERSION
    LHAPDF::setVerbosity((LHAPDF::Verbosity)v);
#endif
    LHAPDF::initPDFSet(PDFset, PDFsubset);
  }

  void Resummation::SetMtop(double mt) {
    mtop = mt;
  }
  void Resummation::SetMbottom(double mb) {
    mbot = mb;
  }
  void Resummation::SetMcharm(double mc) {
    mcha = mc;
  }
  void Resummation::Setmuf(double mu) {
    muF = mu;
  }
  void Resummation::UseLargeMtopEFT(bool use) {
    mtEFT = use;
  }

  void Resummation::SetPrescription(prescription p, bool exponentiate_g0, bool resumDistrConst, bool expBorel) {
    prescr = p;
    if(p < 0) SetMPpars();
    else SetBPpars();
    //shift = PSfactor;
    g0exp = exponentiate_g0;
    expDconst = resumDistrConst;
    if(expBorel && prescr>0) expBP=true;
  }
  void Resummation::SetBPpars(unsigned int Ncheb, double cutoff) {
    if(expBP) SetMPpars(Ncheb);
    Nchebyshev = Ncheb;
    BPcutoff = cutoff;
  }
  //void Resummation::SetMPpars(unsigned int Ncheb, MPmeth method, double beta, int gamma, double z_min) {
  void Resummation::SetMPpars(unsigned int Ncheb, double beta, int gamma, double z_min) {
    Nchebyshev = Ncheb;
    //MPmethod = method;
    MPbeta = beta;
    MPgamma = gamma;
    MPzmin = z_min;
  }
  void Resummation::SetMPpars(unsigned int Ncheb) {
    SetMPpars(Ncheb, 1., 0, 1.e-3);
    //Nchebyshev = Ncheb;
    //MPbeta  = DefaultBeta();
    //MPgamma = DefaultGamma();
    //MPzmin = 1.e-4;
  }
  void Resummation::SetAPimprovement(APtype impr) {
    APimpr = impr;
  }


  double Resummation::xsLL   (double m, double sqrt_s, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    return res_xs(m, sqrt_s, muF_m_ratio, muR_m_ratio,    LL, match_order, asinput);
  }
  double Resummation::xsNLL  (double m, double sqrt_s, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    return res_xs(m, sqrt_s, muF_m_ratio, muR_m_ratio,   NLL, match_order, asinput);
  }
  double Resummation::xsNNLL (double m, double sqrt_s, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    return res_xs(m, sqrt_s, muF_m_ratio, muR_m_ratio,  NNLL, match_order, asinput);
  }
  double Resummation::xsNNNLL(double m, double sqrt_s, double muF_m_ratio, double muR_m_ratio, order match_order, double *asinput) {
    return res_xs(m, sqrt_s, muF_m_ratio, muR_m_ratio, NNNLL, match_order, asinput);
  }








  ////////////////////////////////
  //      NON-CLASS  PART       //
  ////////////////////////////////

  struct s_lum  { double x, mu, M; process p; collider c; };

  // for total cross-section
  double Lum_integrand(double z, void* par) {
    s_lum pars = *(s_lum *) par;
    double x   = pars.x;
    double mu  = pars.mu;
    double M   = pars.M;
    process p  = pars.p;
    collider c = pars.c;
    double lum;
    if(p == ggH) lum = lumi::lumi_gg(z,x/z,mu)/z;
    else lum = lumi::lumi_DY(z,x/z,mu,M,p,c)/z;
    return lum;
  }
  double Lum_inclusive(double x, double mu, double M, process p, collider c) {
    double precision = 1e-4;
    s_lum pars = {x, mu, M, p, c};
    return integration::gauss(Lum_integrand,x,1.,precision,NULL,&pars);
  }
  double Lum(double x, Resummation res) {
    return Lum_inclusive(x, res.muF, res.M, res.proc, res.coll);
  }













  ////////////////////////////////
  //       PRIVATE  PART        //
  ////////////////////////////////

  // default threshold masses, Z mass and as(mZ)
  //const double def_mass_c=1.27, def_mass_b=4.20, def_mass_t=172.5;
  //const double def_mass_Z = 91.18, def_as_Z = 0.118;
  //
  void Resummation::init_mass_thresholds(double mc, double mb, double mt) {
    mass_c = mc;
    mass_b = mb;
    mass_t = mt;
    /*
    cout << endl << "Using threshold masses from PDF set:" << endl
	 << "m_c = " << mass_c << " GeV" << endl
	 << "m_b = " << mass_b << " GeV" << endl
	 << "m_t = " << mass_t << " GeV" << endl
	 << endl;
    */
  }
  int Resummation::nf_of_mu(double mu) {
    int res = 0;
    if(mu>=mass_t) res=6;
    else if(mu<mass_t && mu>=mass_b) res=5;
    else if(mu<mass_b && mu>=mass_c) res=4;
    else res=3;
    return res;
  }

  double Resummation::DefaultBeta() {
    if(proc==ggH)
      return 3.;
    else //if(proc==DYneutral)
      return 2.;
  }
  int Resummation::DefaultGamma() {
    if(proc==ggH)
      return 9;
    else //if(proc==DYneutral)
      return 7;
  }

  // Landau pole
  double Resummation::Nlandau() {
    double a = 0;
    if(APimpr == AP1) a = 1;
    if(APimpr == AP2) a = 2;
    if(APimpr == AP && proc == ggH) a = 3;
    if(abs(prescr) != 10) a -= 0.5; // when prescr isn't MP00 or BP00
    //
    return exp(1./ab) - a;
  }


  double Resummation::res_xs(double m, double sqrt_s,
			     double muF_m_ratio, double muR_m_ratio,
			     log_order LOGo, order match_order, double *asinput) {
    //
    tau = m*m/sqrt_s/sqrt_s;
    M   = m;
    muF = m*muF_m_ratio;
    muR = m*muR_m_ratio;
    nf = nf_of_mu(m); ///// muF, muR or m ?????????
    if(!(last_nf==nf)) {
      b0 = QCD::b0(nf);
      b1 = QCD::b1(nf);
      b2 = QCD::b2(nf);
      b3 = QCD::b3(nf);
    }
    if(asinput == NULL) as = LHAPDF::alphasPDF(muR); // /50.;
    else as = *asinput; // /50.;
    //cout << as << endl;
    ab = 2*b0*as;
    InitResummCoeffs();  // requires nf and m -> leave it here
    LL_order = LOGo;
    matching_order = match_order;
    //
    //gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    //gsl_set_error_handler_off();

    Chebyshev cheb(this);
    bool redo_cheb = !(last_M==M && last_muFrat==muF_m_ratio && last_sqrts==sqrt_s && last_Ncheb==Nchebyshev);
    if(redo_cheb) {
      luminosity_in_tau = Lum(tau, *this);
      if(!firstinit) delete[] coeff;
      coeff = new double[Nchebyshev+1];
    }
    double result = 0.;

    if(prescr < 0 || expBP) { // then it's a MP
      MP_slope = 1.;
      //if(MPmethod == Log) MPbeta = 1;
      MP_reN0 = MPbeta + 3.;
      double NL = Nlandau();
      if(MP_reN0 >= NL) {
	MP_reN0 = ( MPbeta + NL )/2.;
	if(MP_reN0 >= NL) {
	  cout << "ERROR: Landau pole N_L = " << NL
	       << " to the left of the rightmost singularity: " << MPbeta << endl;
	  exit(-13);
	}
      }
      /*
      if(MPmethod == Manip)
	cheb.ApproxLuminosity(coeff, Nchebyshev, MPbeta, MPgamma);
      else if(MPmethod == Log) {
	double zmin = 1.e-6;
	cheb.ApproxLuminosity_log(coeff, Nchebyshev, zmin);
      */
      if(tau<MPzmin) MPzmin=tau;
      if(redo_cheb) cheb.ApproxLuminosity(coeff, Nchebyshev, MPbeta, MPgamma, MPzmin);
      result = MinimalPrescription()/luminosity_in_tau;
    }
    else if(prescr == BP0 || prescr == BP00) {
      if(redo_cheb) cheb.ApproxLumPlus_log1z(coeff, tau, Nchebyshev);
      result = BorelPrescription();
    }
    else {
      if(redo_cheb) cheb.ApproxLumPlus(coeff, tau, Nchebyshev);
      result = BorelPrescription();
    }
    //gsl_set_error_handler (old_handler);
    //
    last_muFrat = muF_m_ratio;
    last_muRrat = muR_m_ratio;
    last_M      = M;
    last_sqrts  = sqrt_s;
    last_nf     = nf;
    last_mtop   = mtop;
    last_EFT    = mtEFT;
    last_Ncheb  = Nchebyshev;
    firstinit = false;
    //
    return result;
  }






#include "math/mellin_Dk.cc"
  dcomplex Resummation::mellD(int k, dcomplex N) {
    if(expDconst) {
      if(abs(prescr) == 1)
	return mell_D(k,N);
      else if(abs(prescr) == 2)
	return mell_DB(k,N);
      else if(abs(prescr) == 10)
	return mell_DMconst(k,N);
      else if(abs(prescr) == 11)
	return mell_DMconst(k,N-0.5);
    }
    else {
      if(abs(prescr) == 1)
	return mell_D_noconst(k,N);
      else if(abs(prescr) == 2)
	return mell_DB_noconst(k,N);
      else if(abs(prescr) == 10)
	return mell_DM(k,N);
      else if(abs(prescr) == 11)
	return mell_DM(k,N-0.5);
    }
    return 0;
  }


  dcomplex Resummation::APimprovement(dcomplex (Resummation::*f)(dcomplex,void*), dcomplex N, void *p) {
    dcomplex res;
    if(proc == ggH) {
      if(APimpr == AP0 ) res =   (this->*f)(N,p);
      if(APimpr == AP1 ) res =   (this->*f)(N+1,p);
      if(APimpr == AP2 ) res = 2*(this->*f)(N,p) - 3*(this->*f)(N+1,p) + 2*(this->*f)(N+2,p);
      if(APimpr == AP12) res =   (this->*f)(N,p) -   (this->*f)(N+1,p) +   (this->*f)(N+2,p);
      if(APimpr == AP  ) res =   (this->*f)(N-1,p) - 2*(this->*f)(N,p) + 3*(this->*f)(N+1,p) - 2*(this->*f)(N+2,p) + (this->*f)(N+3,p);
    }
    else { // DY
      if(APimpr == AP0 ) res = (this->*f)(N,p);
      if(APimpr == AP1 ) res = (this->*f)(N+1,p);
      if(APimpr == AP2 ) res = (this->*f)(N,p)/2. + (this->*f)(N+2,p)/2.;
      if(APimpr == AP12) res = (this->*f)(N,p)/4. + (this->*f)(N+2,p)/4. + (this->*f)(N+1,p)/2.;
      if(APimpr == AP  ) res = (this->*f)(N,p)/2. + (this->*f)(N+2,p)/2.;
    }
    return res;
  }


  inline dcomplex NofZ(dcomplex z, double ab) {
    return exp(-z/ab);
  }
  inline dcomplex ZofN(dcomplex N, double ab) {
    return -ab*log(N);
  }



  // Sudakov exponent
  dcomplex Resummation::S(dcomplex N, void *p) {
    dcomplex z = ZofN(N,ab);
    dcomplex res = g1(z)/ab;
    if(LL_order > 0) res += g2(z);
    if(LL_order > 1) res += g3(z) *as;
    if(LL_order > 2) res += g4(z) *as*as;
    return res;
  }
  double Resummation::g0() {
    double value = 1.;
    if(g0exp) {
      if(LL_order > 0) value = exp(g01*as);
      if(LL_order > 1) value = exp(g01*as + (g02-g01*g01/2.)*as*as);
      if(LL_order > 2) value = exp(g01*as + (g02-g01*g01/2.)*as*as + (g03-g01*g02+g01*g01*g01/3.)*as*as*as);
    } else {
      if(LL_order > 0) value += g01 * as;
      if(LL_order > 1) value += g02 * as*as;
      if(LL_order > 2) value += g03 * as*as*as;
    }
    return value;
  }
  double Resummation::g0sub() {
    double value = 0;
    if(LL_order > 0) value += g0sub1 * as;
    if(LL_order > 1) value += g0sub2 * as*as;
    if(LL_order > 2) value += g0sub3 * as*as*as;
    return value;
  }
  //
  //
  dcomplex Resummation::asS1(dcomplex N, void *p) {
    dcomplex L = ZofN(N,ab)/ab;
    dcomplex value;
    dcomplex L1 = L;
    dcomplex L2 = L*L;
    if(expBP) {
      L1 = mellD(0,N);
      L2 = 2*mellD(1,N) + 2*EulerGamma*L1;
    }
    value = ab*L2*A1/b0;
    if(LL_order > 0) value += ab*L1*2*(log(M/muF)-EulerGamma)*A1/b0;
    return value;
  }
  dcomplex Resummation::as2S2(dcomplex N, void *p) {
    dcomplex L = ZofN(N,ab)/ab;
    dcomplex value;
    dcomplex L1 = L;
    dcomplex L2 = L*L;
    dcomplex L3 = L*L*L;
    if(expBP) {
      L1 = mellD(0,N);
      L2 = 2*mellD(1,N) + 2*EulerGamma*L1;
      L3 = 3*mellD(2,N)-3*ZETA2*L1 + 3*EulerGamma*L2 - 3*EulerGamma*EulerGamma*L1;
    }
    value = - ab*ab*L3*A1/b0/3.;
    if(LL_order > 0) value += ab*ab*L2* ( A1*(EulerGamma-log(M/muR))/b0 + A2/2./b0/b0 );
    if(LL_order > 1) value += ab*ab*L1* ( D2/2./b0/b0 - 2.*A2*EulerGamma/b0/b0 - 2.*A1*(EulerGamma*EulerGamma+ZETA2)/b0
					  + 4.*A1*EulerGamma*log(M/muR)/b0 +2.*A2*log(M/muF)/b0/b0
					  - 2.*A1*(pow(log(M/muR),2.)-pow(log(muF/muR),2.))/b0 )/2.;
    return value;
  }
  //
  dcomplex Resummation::expS(dcomplex N) {
    if(expBP) return exp( BorelPrescriptionExp(N) );
    return exp( APimprovement(&Resummation::S,N) );
  }
  dcomplex Resummation::g0expS_order1(dcomplex N) {
    dcomplex value;
    value = APimprovement(&Resummation::asS1,N);
    if(LL_order > 0) value += g01*as;
    return value;
  }
  dcomplex Resummation::g0expS_order2(dcomplex N) {
    dcomplex value;
    value = pow(APimprovement(&Resummation::asS1,N),2.)/2. + APimprovement(&Resummation::as2S2,N);
    if(LL_order > 0) value += g01*as*APimprovement(&Resummation::asS1,N);
    if(LL_order > 1) value += g02*as*as;
    return value;
  }
  //
  dcomplex Resummation::g0expS_matched(dcomplex z) {
    dcomplex N = NofZ(z,ab);
    dcomplex value;
    //
    value = expS(N) * g0() - g0sub() - 1.;
    //
    if(matching_order > 0) value -= g0expS_order1(N) - g0sub1*as;
    if(matching_order > 1) value -= g0expS_order2(N) - g0sub2*as*as;
    return value;
  }






  ///////////////////////////
  //  BOREL  PRESCRIPTION  //
  ///////////////////////////

  dcomplex Resummation::cheb_term(dcomplex xi, int j) {
    dcomplex val;
    //if     (prescr == BP00) val = TruncatedGamma(xi+j+1, (shift?3./2.:1.)*log(1./tau)) * (shift?exp((xi+j+1)*log(2./3.)):1.);
    //else if(prescr == BP1 ) val = (shift ? IncompleteBeta(xi+j+1, 2., 1.-tau) : exp((xi+j+1.)*log(1.-tau))/(xi+j+1.) );
    //else if(prescr == BP2 ) val = IncompleteBeta(xi+j+1, 1.-xi/2.+(shift?1.:0.), 1.-tau);
    if     (prescr == BP00) val = TruncatedGamma(xi+j+1,   log(1./tau));
    else if(prescr == BP0 ) val = TruncatedGamma(xi+j+1, 2*log(1./tau)) * exp((xi+j+1)*log(2.));
    else if(prescr == BP1 ) val = exp((xi+j+1.)*log(1.-tau))/(xi+j+1.);
    else if(prescr == BP2 ) val = IncompleteBeta(xi+j+1, 1.-xi/2., 1.-tau);
    return val;
  }
  dcomplex Resummation::delta_term(dcomplex xi) {
    dcomplex val = cheb_term(xi,-1);
    /*
    //if     (prescr == BP00) val = TruncatedGamma(xi, (shift?3./2.:1.)*log(1./tau)) * (shift?exp(xi*log(2./3.)):1.);
    //else if(prescr == BP1 ) val = (shift ? IncompleteBeta(xi, 2., 1.-tau) : exp((xi)*log(1-tau))/xi);
    //else if(prescr == BP2 ) val = IncompleteBeta(xi, 1.-xi/2.+(shift?1.:0.), 1.-tau);
    if     (prescr == BP00) val = TruncatedGamma(xi,   log(1./tau));
    else if(prescr == BP0 ) val = TruncatedGamma(xi, 2*log(1./tau)) * exp(xi*log(2.));
    else if(prescr == BP1 ) val = exp((xi)*log(1-tau))/xi;
    else if(prescr == BP2 ) val = IncompleteBeta(xi, 1.-xi/2., 1.-tau);
    */
    if(expDconst) val -= 1./xi;
    return val;
  }
  dcomplex Resummation::convolution(dcomplex xi) { 
    dcomplex value = 0.;
    for(unsigned int j=0; j<=Nchebyshev; j++) {
      value += coeff[j] * cheb_term(xi,j);
    }
    //return value + luminosity_in_tau * delta_term(xi);
    return value + delta_term(xi);
  }
  dcomplex Resummation::BorelIntegrand(double w, dcomplex xi) {
    dcomplex result = g0expS_matched(w/xi)*exp(-w/ab) * convolution(xi) /Gamma(xi+1)/ab; // /2./M_PI
    return result;
  }
  double Resummation::BorelIntegrand_chvar(double x1, double x2) {
    double w = x1*BPcutoff;
    //
    double NL = Nlandau();
    //double cutlength = BPcutoff/ab/log(NL);
    double cutlength = w/ab/log(NL);
    //
    double r = 1.3 + 2*cutlength;
    dcomplex xi = r * exp(I*x2*M_PI);
    double result = real(xi*BorelIntegrand(w, xi)) * BPcutoff;  // *2.*M_PI;
    //double r = 1. + 2*cutlength;
    //dcomplex xi = r * exp(I*x2*M_PI) - 0.5*cutlength;
    //double result = real((xi+0.5*cutlength)*BorelIntegrand(w, xi)) * BPcutoff;  // *2.*M_PI;
    //
    //cout << result << endl;
    return result;
  }
  int integrandCUBA(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    Resummation Bres = *(Resummation *) pars;
    res[0] = Bres.BorelIntegrand_chvar(x[0],x[1]);
    return 0;
  }
  double Resummation::BorelPrescription() {
    double cres[1], error[1], prob[1];
    double epsrel=1.e-3, epsabs=1.e-100;
    int last = 4;
    int verbose = 0;
    int nregions, neval, fail;
    Cuhre(2, 1, integrandCUBA, this,
	  epsrel, epsabs, verbose | last,
	  0, 500000, 9,
	  &nregions, &neval, &fail, cres, error, prob);
    if(fail == -1) {
      cout << "  ERROR: dimension out of range." << endl;
    } else if(fail>0) {
      cout << "  ERROR: precision not reached." << endl;
    }
    return cres[0];
  }






  ////////////////////////////////////////
  // Borel Prescription at the exponent //
  ////////////////////////////////////////

  dcomplex Resummation::NLogGen(dcomplex N, void *par) {
    dcomplex res;
    dcomplex xi = *(dcomplex *)par;
    if     (prescr == BP00) res = exp(-xi*log(N));
    else if(prescr == BP0 ) res = exp(-xi*log(N-0.5));
    else if(prescr == BP1 ) res = Gamma(N)/Gamma(N+xi);
    else if(prescr == BP2 ) res = Gamma(N-xi/2.)/Gamma(N+xi/2.);
    if(expDconst) res -= 1./Gamma(1+xi);
    //if(expDconst) res -= Gamma(1-xi/2.)/Gamma(1+xi/2.);
    return res;
  }
  dcomplex Resummation::BorelIntegrandExp(double w, dcomplex xi, dcomplex N) {
    dcomplex result = S(NofZ(w/xi,ab),NULL)*exp(-w/ab) * APimprovement(&Resummation::NLogGen,N,&xi) /xi/ab; // /2./M_PI
    return result;
  }
  dcomplex Resummation::BorelIntegrandExp_chvar(double x1, double x2, dcomplex N) {
    double w = x1*BPcutoff;
    //
    double NL = Nlandau();
    //double cutlength = BPcutoff/ab/log(NL);
    double cutlength = w/ab/log(NL);
    //
    double r = 1.3 + 2*cutlength;
    //double r = 0.7 + 2*cutlength;
    //double r = 1.3 + 1.5*cutlength;
    //double r = 1.5*cutlength;
    dcomplex xi = r * exp(I*x2*2*M_PI);
    dcomplex result = xi*BorelIntegrandExp(w, xi, N) * BPcutoff;  // *2.*M_PI;
    //double r = 1. + 2*cutlength;
    //dcomplex xi = r * exp(I*x2*2*M_PI) - 0.5*cutlength;
    //double result = real((xi+0.5*cutlength)*BorelIntegrand(w, xi)) * BPcutoff;  // *2.*M_PI;
    //
    //cout << result << endl;
    return result;
  }
  struct BPexpintpars {
    Resummation *res;
    dcomplex N;
  };
  int integrandCUBAexp(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
    BPexpintpars p = *(BPexpintpars *) pars;
    Resummation *Bres = p.res;
    dcomplex N = p.N;
    dcomplex r = Bres->BorelIntegrandExp_chvar(x[0],x[1],N);
    res[0] = real(r);
    res[1] = imag(r);
    return 0;
  }
  dcomplex Resummation::BorelPrescriptionExp(dcomplex N) {
    double cres[2], error[2], prob[2];
    double epsrel=2.e-2, epsabs=1.e-100;
    //double epsrel=5.e-3, epsabs=1.e-100;
    int last = 4;
    int verbose = 0;
    int nregions, neval, fail;
    nregions=1; // to remove the warning
    int mineval = 0, maxeval = 40000000;
    BPexpintpars pars = { this, N };
    Cuhre(2, 2, integrandCUBAexp, &pars, epsrel, epsabs, verbose | last, mineval, maxeval, 13, &nregions, &neval, &fail, cres, error, prob);
    //Suave(2, 2, integrandCUBAexp, &pars, epsrel, epsabs, verbose | last, 0, mineval, maxeval, 1000, 25., &nregions, &neval, &fail, cres, error, prob);
    //Vegas(2, 2, integrandCUBAexp, &pars, epsrel, epsabs, verbose | last, 0, mineval, maxeval, 10000, 2000, 10000, 1, NULL, &neval, &fail, cres, error, prob);
    if(fail == -1) {
      cout << "  ERROR: dimension out of range." << endl;
    } else if(fail>0) {
      cout << "  ERROR: precision not reached." << endl;
    }
    return dcomplex(cres[0],cres[1]);
  }






  ///////////////////////////
  //  MINIMAL PRESCRIPTION //
  ///////////////////////////

  dcomplex Resummation::LuminosityN(dcomplex N){
    dcomplex value = 0;
    double zmax = 1.; // usless now: relevant only for rapidity distributions
    for(unsigned int i=0; i<=Nchebyshev; i++) {
      for(int j=0; j<=MPgamma; j++) {
	value += coeff[i] * pow(-1.,j) * gsl_sf_choose(MPgamma,j) / pow(N+j-MPbeta,i+1.);
      }
    }
    value *= pow(zmax,N-MPbeta+MPgamma);
    /*
      if(MPmethod==0) // total
	if(isINT(MPgamma)) {
	  for(int j=0; j<=MPgamma; j++) {
	    value += coeff[i] * pow(-1.,j) * gsl_sf_choose(INT(MPgamma),j) / (N+i+j-MPbeta);
	  }
	} else {
	  value += exp(log(Gamma(N+i-MPbeta)) - log(Gamma(N + i + 1 - MPbeta + MPgamma)))
	    * gsl_sf_gamma(1+MPgamma) * coeff[i];
	}
      else if(MPmethod==1) // total log
	value += coeff[i]/pow(N-1.,i+1.);
      else if(MPmethod==0) // rap
	if(isINT(MPgamma)) {
	  for(int j=0; j<=MPgamma; j++) {
	    value += coeff_cheb_lum[i] * pow(-1.,j) * gsl_sf_choose(INT(MPgamma),j)
              * pow(em2Y, i+j-MPbeta+MPgamma-j) / (N+i+j-MPbeta);
	  }
	} else {
	  cerr << "Not yet implemented for non-integer values of MPgamma" << endl;
	  exit(-33);
	  //value += IncompleteBeta(N+i-MPbeta, 1.+MPgamma, em2Y) * coeff_cheb_lum[i];
	}
      else if(MPmethod==1) // rap log
	value += coeff_cheb_lum[i]/pow(N-1.,i+1.)/em2Y;
      else {
	cerr << "Non existent method for luminosity Chebyshev approximation." << endl;
	exit(-12);
      }
    }
    */
    return value;
  }
  dcomplex Resummation::lum_times_g0expS_matched(dcomplex N) {
    dcomplex z;
    if(prescr == MP00) z = -ab*log(N);      // original MP
    if(prescr == MP0 ) z = -ab*log(N-0.5);  // removes spurious sqrt(z)
    if(prescr == MP2 ) z = -ab*psi(N);      // also restores more physical logs (correct analytic structure)
    if(expBP && prescr>0) z = ZofN(N,ab);
    return g0expS_matched(z) * LuminosityN(N);
  }
  double Minimal_integrand(double t, void *p) {
    Resummation Mres = *(Resummation *) p;
    double x = Mres.tau;
    dcomplex N = Mres.MP_reN0 + (Mres.MP_slope-I)*log(t);
    dcomplex res = (I-Mres.MP_slope) * exp(-N*log(x)) * Mres.lum_times_g0expS_matched(N)/t;
    return imag(res)/M_PI;
  }
  /*
  double Minimal_integrand_inf(double t, void *p) {
    Resummation Mres = *(Resummation *) p;
    double x = Mres.tau;
    dcomplex N = Mres.MP_reN0 + (I-Mres.MP_slope)*t;
    dcomplex res = (I-Mres.MP_slope) * exp(-N*log(x)) * Mres.lum_times_g0expS_matched(N);
    return imag(res)/M_PI;
  }
  */
  dcomplex Resummation::InverseMellinIntegrand(dcomplex N) {
    return exp( -N*log(tau) ) * lum_times_g0expS_matched(N) / M_PI;
  }
  double Resummation::InverseMellin() { // Talbot algorithm
    if(tau==1.) return 0;
    //double log1ox = -log(tau);
    double res = 0.; 
    // prec ~ 10^{-0.6*MM}
    int MM = 18; // tune
    //double r = 1.+2.*MM/5./log1ox;
    double r = 2.5; // tune
    double t, cott, sigma;
    dcomplex N;
    res = real(0.5 * InverseMellinIntegrand(r));
    for(int k=1; k<MM; k++) {
      t = k*M_PI/MM;
      cott = 1./tan(t);
      sigma = t + cott*(t*cott - 1.);
      N = r*t*(cott+I);
      res += real(InverseMellinIntegrand(N) * (1.+I*sigma));
    }
    res *= M_PI*r/MM;
    return res;
  }
  double Resummation::MinimalPrescription() {
    //return InverseMellin();
    double res=0, prec = 1.e-3;
    if(expBP) prec *= 5;
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    res = integration::gauss(Minimal_integrand,0.,1.,prec,NULL,this);
    //res = integration::gauss_inf(Minimal_integrand_inf,0.,prec,NULL,&x);  // fails..
    gsl_set_error_handler (old_handler);
    return res;
  }

  /*
  double rap_Minimal(double Y, double x) {
    em2Y = exp(-2*ABS(Y));
    double xe2Y = x/em2Y;
    double prec = 1e-4;
    //return integration::dgauss(Minimal_integrand,0.,1.,prec,NULL,&xe2Y);
    return integration::gauss_inf(Minimal_integrand_inf,0.,prec,NULL,&xe2Y);  // faster
  }
  */










  ///////////////////////////
  //  AUXILIARY FUNCTIONS  //
  ///////////////////////////



  // functions in the exponent; gk means N^(k-1)LL approximation
  // z is the variable ab*log(1/N) = 2 as b0 log(1/N)
  dcomplex Resummation::g1(dcomplex z) {
    return (log(z+1)*(z+1)-z)*2.*A1/b0;
  }
  dcomplex Resummation::g2(dcomplex z) {
    return (( pow(log(z+1),2)*0.5 - z + log(z+1) ) * ( A1*b1/b0/b0 ) +
	    (z - log(z+1) ) * A2/b0/b0
	    - (log(z+1) * (EulerGamma - log(M/muR)) + z*log(muF/muR) )*2.*A1/b0 );
  }
  // g3 taken from Moch-Vermaseren-Vogt, hep-ph/0506288
  dcomplex Resummation::g3(dcomplex z) {
    return ( (z*z/(z+1.)) * (A3 - A1*b2 + A1*b1*b1 - A2*b1)/2./b0/b0/b0
	     + ( log(z+1.)/(z+1.) + pow(log(z+1.),2.)/(z+1.)/2. ) * A1*b1*b1/b0/b0/b0
	     + log(z+1.) * A1*(b2-b1*b1)/b0/b0/b0
	     + ( z/(z+1.) - log(z+1.)/(z+1.) ) * ( 2.*A1*b1*EulerGamma + A2*b1/b0 )/b0/b0
	     - ( z/(z+1.) ) * ( A1*b2/b0/b0 + 2.*A1*(EulerGamma*EulerGamma + ZETA2) + 2.*A2*EulerGamma/b0 )/b0
	     + ( z/(z+1.) ) * D2/2./b0/b0
	     + ( ( z/(z+1.) ) * ( 2.*A1*EulerGamma + A2/b0 -A1*b1/b0 )/b0
	      + log(z+1.)/(z+1.) * A1*b1/b0/b0 ) * 2.*log(M/muR)
	     - z * A2/b0/b0 * 2.*log(muF/muR)
	     + ( z * pow(log(muF/muR),2.) - z/(z+1.) * pow(log(M/muR),2.) ) * 2.*A1/b0 
	     )*b0;
  }
  //
  // g4 taken from Moch-Vermaseren-Vogt, hep-ph/0506288 (translated from tex to C++ with Mathematica)
  template<class T> T Power(T x, unsigned int n) {
    T res = 1;
    for(unsigned int i=0; i<n; i++) res *= x;
    return res;
  }
  dcomplex Resummation::g4(dcomplex z) {
    return -(-6*A3*b1*z + 6*A2*Power(b1,2)*z - 6*A1*b1*b2*z + 6*A1*b3*z + 6*b0*b1*D2*z - 
      6*b0*D3*z + 24*A3*b0*EulerGamma*z - 24*A2*b0*b1*EulerGamma*z - 
      24*Power(b0,2)*D2*EulerGamma*z + 48*A2*Power(b0,2)*Power(EulerGamma,2)*z + 
      32*A1*Power(b0,3)*Power(EulerGamma,3)*z - 6*A4*Power(z,2) + 3*A3*b1*Power(z,2) + 
      3*A2*Power(b1,2)*Power(z,2) - 9*A1*b1*b2*Power(z,2) + 9*A1*b3*Power(z,2) + 
      3*b0*b1*D2*Power(z,2) - 3*b0*D3*Power(z,2) + 12*A3*b0*EulerGamma*Power(z,2) - 
      12*A2*b0*b1*EulerGamma*Power(z,2) + 12*A1*b0*Power(b1,2)*EulerGamma*Power(z,2) - 
      12*A1*b0*b2*EulerGamma*Power(z,2) - 12*Power(b0,2)*D2*EulerGamma*Power(z,2) + 
      24*A2*Power(b0,2)*Power(EulerGamma,2)*Power(z,2) + 
      16*A1*Power(b0,3)*Power(EulerGamma,3)*Power(z,2) - 4*A4*Power(z,3) + 
      4*A3*b1*Power(z,3) - 4*A2*Power(b1,2)*Power(z,3) + 4*A1*Power(b1,3)*Power(z,3) + 
      4*A2*b2*Power(z,3) - 8*A1*b1*b2*Power(z,3) + 4*A1*b3*Power(z,3) + 
      48*A2*Power(b0,2)*z*ZETA2 + 96*A1*Power(b0,3)*EulerGamma*z*ZETA2 + 
      24*A2*Power(b0,2)*Power(z,2)*ZETA2 + 
      48*A1*Power(b0,3)*EulerGamma*Power(z,2)*ZETA2 + 64*A1*Power(b0,3)*z*ZETA3 + 
      32*A1*Power(b0,3)*Power(z,2)*ZETA3 - 
      16*A1*Power(b0,3)*z*(2 + z)*Power(log(M/muR),3) + 
      24*A3*b0*z*Power(1 + z,2)*log(muF/muR) - 
      24*Power(b0,2)*(2*A2 + A1*b1)*z*Power(1 + z,2)*Power(log(muF/muR),2) + 
      32*A1*Power(b0,3)*z*Power(1 + z,2)*Power(log(muF/muR),3) + 6*A3*b1*log(1 + z) - 
      6*A2*Power(b1,2)*log(1 + z) + 6*A1*b1*b2*log(1 + z) - 6*A1*b3*log(1 + z) - 
      6*b0*b1*D2*log(1 + z) + 24*A2*b0*b1*EulerGamma*log(1 + z) + 
      24*A1*Power(b0,2)*b1*Power(EulerGamma,2)*log(1 + z) + 12*A1*b1*b2*z*log(1 + z) - 
      12*A1*b3*z*log(1 + z) - 6*A1*Power(b1,3)*Power(z,2)*log(1 + z) + 
      12*A1*b1*b2*Power(z,2)*log(1 + z) - 6*A1*b3*Power(z,2)*log(1 + z) + 
      24*A1*Power(b0,2)*b1*ZETA2*log(1 + z) - 6*A2*Power(b1,2)*Power(log(1 + z),2) - 
      12*A1*b0*Power(b1,2)*EulerGamma*Power(log(1 + z),2) + 
      2*A1*Power(b1,3)*Power(log(1 + z),3) + 
      12*Power(b0,2)*Power(log(M/muR),2)*
       (2*(A2 + 2*A1*b0*EulerGamma)*z*(2 + z) + 2*A1*b1*log(1 + z)) - 
      12*b0*log(M/muR)*(z*(-2*b0*D2 + 8*A1*Power(b0,2)*Power(EulerGamma,2) + 
            A1*Power(b1,2)*z - A1*b2*z - b0*D2*z + 
            4*A1*Power(b0,2)*Power(EulerGamma,2)*z + A3*(2 + z) - 
            A2*(b1 - 4*b0*EulerGamma)*(2 + z) + 8*A1*Power(b0,2)*ZETA2 + 
            4*A1*Power(b0,2)*z*ZETA2) + 2*b1*(A2 + 2*A1*b0*EulerGamma)*log(1 + z) - 
			A1*Power(b1,2)*Power(log(1 + z),2)))/(12.*Power(b0,2)*Power(1 + z,2));
  }



  void Resummation::InitResummCoeffs() {
    if(proc >= 1) {  // Drell-Yan
      if(!(last_nf==nf)) InitCoeffDY();
      InitConstDY();
    } else if(proc == ggH) {  // Higgs
      if(!(last_nf==nf)) InitCoeffHiggs();
      InitConstHiggs();
    }
    return;
  }

  void Resummation::InitCoeffDY() {
    using namespace QCD;
    A1 = CF/M_PI;
    A2 = CF/2.* (67.*CA/18.-5./9.*nf-CA*ZETA2) /M_PI/M_PI;
    A3 = ( CA*CA* (245./24.-67.*ZETA2/9.+11.*ZETA3/6.+11.*ZETA2*ZETA2/5.) +
	   ( (2.*ZETA3 - 55./24.)*CF + (10.*ZETA2/9.-7.*ZETA3/3.-209./108.)*CA )*nf
	   - nf*nf*1./27. ) *CF/4. /M_PI/M_PI/M_PI;
    A4 = 0; // use Pade' [1/1]
    if(nf == 3) A4 = 0.314757;
    if(nf == 4) A4 = 0.172958;
    if(nf == 5) A4 = 0.0622776;
    D2 = CF * ( CA * (176.*ZETA2/3. + 56.*ZETA3 - 1616./27.) + nf*(224./27.-32.*ZETA2/3.) )/16./M_PI/M_PI; // checked: OK.
    D3 = CF * ( CA*CA *(-594058./729. +98224./81.*ZETA2 +40144./27.*ZETA3 -2992./15.*ZETA2*ZETA2 -352./3.*ZETA2*ZETA3 -384.*ZETA5 )
		+CA*nf*( 125252./729. -29392./81.*ZETA2 -2480./9.*ZETA3 +736./15.*ZETA2*ZETA2 )
		+CF*nf*( 3422./27. - 32.*ZETA2 - 608./9.*ZETA3 - 64./5.*ZETA2*ZETA2 )
		+nf*nf*( 320./27.*ZETA3 + 640./27.*ZETA2 - 3712./729.)) /64./M_PI/M_PI/M_PI; // checked: OK.
    return;
  }
  void Resummation::InitCoeffHiggs() {
    using namespace QCD;
    InitCoeffDY();
    A1 *= CA/CF;
    A2 *= CA/CF;
    A3 *= CA/CF;
    A4 *= CA/CF;
    D2 *= CA/CF;
    D3 *= CA/CF;
    //
    // prints g4 at order lambda (I used it for ggHiggs, for the NNLL curve)
    // cout << endl << "b0: " << b0*4*M_PI;
    // cout << endl << "b1: " << beta1*4*M_PI*4*M_PI;
    // cout << endl << "A1: " << A1*4*M_PI;
    // cout << endl << "A2: " << A2*4*M_PI*4*M_PI;
    // cout << endl << "A3: " << A3*4*M_PI*4*M_PI*4*M_PI;
    // cout << endl << "D2: " << D2*4*M_PI*4*M_PI;
    // cout << endl << "D3: " << setprecision(7) << D3*4*M_PI*4*M_PI*4*M_PI;
    // cout << endl << "g:  " << EulerGamma;
    // cout << endl << "z2: " << ZETA2;
    // cout << endl << "z3: " << ZETA3;
    // cout << endl << -D3 + (4*A3 - 4*D2*b0)*EulerGamma  + (8*A2*b0 + 4*A1*b1*b0) * (EulerGamma*EulerGamma + ZETA2)
    //   + 16./3.*A1*b0*b0* (3*EulerGamma*ZETA2 + EulerGamma*EulerGamma*EulerGamma + 2*ZETA3) << endl;
    // b0 *= 4*M_PI;
    // b1 *= 4*M_PI*4*M_PI;
    // A1 *= 4*M_PI;
    // A2 *= 4*M_PI*4*M_PI;
    // A3 *= 4*M_PI*4*M_PI*4*M_PI;
    // D2 *= 4*M_PI*4*M_PI;
    // D3 *= 4*M_PI*4*M_PI*4*M_PI;
    // double res = -D3 + (4*A3 - 4*D2*b0)*EulerGamma  + (8*A2*b0 + 4*A1*b1*b0) * (EulerGamma*EulerGamma + ZETA2)
    //   + 16./3.*A1*b0*b0* (3*EulerGamma*ZETA2 + EulerGamma*EulerGamma*EulerGamma + 2*ZETA3);
    // cout << endl << res << endl << res/pow(4*M_PI,3) << endl;
    return;
  }
  void Resummation::InitConstDY() {
    using namespace QCD;
    g01 = CF*(4.*ZETA2-4.+2.*EulerGamma*EulerGamma + (4.*EulerGamma-3.)*log(muF/M))/M_PI;
    g02 = ( CF * ( 511./4. - 198.*ZETA2 - 60.*ZETA3 + 552.*ZETA2*ZETA2/5.
		    + 128.*EulerGamma*EulerGamma*(ZETA2-1.+EulerGamma*EulerGamma/4.) )
	    + CA * ( -1535./12. + 376.*ZETA2/3. + 604.*ZETA3/9. - 92.*ZETA2*ZETA2/5. + 1616.*EulerGamma/27.
			   - 56.*EulerGamma*ZETA3 + EulerGamma*EulerGamma* (536./9. - 16.*ZETA2 + 176.*EulerGamma/9.) )
	    + nf * ( 127./6. - 64.*ZETA2/3. + 8.*ZETA3/9. - EulerGamma*(224./27.+EulerGamma*(80.+32.*EulerGamma)/9.) )
	    + 4.*SQR(log(muF/M)) * ( CF * (32.*EulerGamma*EulerGamma - 48.*EulerGamma +18.)
					   + CA * (44.*EulerGamma/3.-11.)
					   + nf * (2.-8.*EulerGamma/3.))
	    - 2.*log(muF/M) * ( CF * (48.*ZETA3 + 72.*ZETA2 -93. -128.*EulerGamma*ZETA2 +128.*EulerGamma
						  + 48.*EulerGamma*EulerGamma -64.*EulerGamma*EulerGamma*EulerGamma)
				       + CA * (193./3. -24.*ZETA3 -88.*ZETA2/3. +16.*EulerGamma*ZETA2 -536*EulerGamma/9.
						-88.*EulerGamma*EulerGamma/3.)
				       + nf * (16.*ZETA2/3. -34./3. +80.*EulerGamma/9. +16.*EulerGamma*EulerGamma/3.))
	    )*CF/16./M_PI/M_PI  - b0*g01*2.*log(muF/muR);
    g03 = 0; // unknown? or known maybe: 1404.0366
    return;
  }


  void Resummation::InitConstHiggs() {
    using namespace QCD;
    //
    double beta0 = b0*M_PI;
    double beta1 = b1*b0*M_PI*M_PI;
    double beta2 = b2*b0*M_PI*M_PI*M_PI;
    //
    if(!(last_nf==nf && last_M==M && last_mtop==mtop && last_EFT==mtEFT)) {
      // EFT
      double PLd1 = 2.*CA * ZETA2;
      double PLd2 = ( 837./16. + 67./2.*ZETA2 - 9./20.*ZETA2*ZETA2 - 165./4.*ZETA3
		      + nf * ( 5./6.*ZETA3 - 5./3.*ZETA2 - 247./36. ) );
      double PLd3 = 1740.02 - 131.928*nf + 1.75719*nf*nf; // new coefficient
      double lHt = 2.*log(M/mtop);
      double W1 = 11./2.;
      double W2 = 121./16. + 2777./144. + 19./8.*lHt + (-67./48. + 2./3.*lHt)*nf;
      double W3 = (-1792967./20736. +  897943./4608.*ZETA3 + 5347./288.*lHt + 209./32.*lHt*lHt
		   + nf * (493./10368. + 209./54.*lHt + 23./16.*lHt*lHt - 110779./6912.*ZETA3)
		   + nf*nf * (-6865./15552. + 77./864.*lHt - lHt*lHt/9. ) );
      d1c = PLd1 + W1;
      d2c = PLd2 + PLd1*W1 + W2;
      d3c = PLd3 + PLd2*W1 + PLd1*W2 + W3;
      //
      if(!mtEFT) {
	initBDV(M, mtop, mbot, mcha);
	d1c = G2l() + 2.*CA * ZETA2;
	if(M < 2*mtop) d2c = gg_nnlo_mt_delta(M, mtop, 1);
	//d3c = ; // not known -> use EFT value
      }
    }
    //
    double d1 = d1c;
    double d2 = d2c;
    double d3 = d3c;
    //
    if(!(last_nf==nf)) {
      gA1 = CA*CA*(67./36.-ZETA2/2.) - 5./18.*CA*nf;
      //gA2 = CA*CA*CA*(245./24.-67./9.*ZETA2+11./6.*ZETA3+11./5.*ZETA2*ZETA2)/4.
      //  + CA*CA*nf*(-209./108.+10./9.*ZETA2-7./3.*ZETA3)/4.
      //  + CA*CF*nf*(-55./24.+2.*ZETA3)/4.
      //  - CA*nf*nf/4./27.;
      gB1 = CA*CA*(2./3.+3./4.*ZETA3) - (CA/6.+CF/8.)*nf;
      gB2 = CA*CA*CA/4.*(79./32.-ZETA2*ZETA3+ZETA2/6.+11./24.*ZETA2*ZETA2+67./6.*ZETA3-5.*ZETA5)
	- CA*CA*nf/4.*(233./288.+ZETA2/6.+ZETA2*ZETA2/12.+5./3.*ZETA3)
	- CA*CF*nf/4.*241./288.
	+ CF*CF*nf/4./16.
	+ nf*nf/4.*(CA*29./288.+CF*11./144.);
    }
    //
    if(!(last_nf==nf && last_M==M && last_mtop==mtop && last_EFT==mtEFT && last_muFrat==muF/M && last_muRrat==muR/M)) {
      double lf  = 2.*log(M/muF);
      double lrf = 2.*log(muR/muF);
      double lr = 2.*log(muR/M);
      /*
      // factorization scale dependence (as=as(muf))
      d3 += (72.*ZETA3+18.*beta0*ZETA2)*lf*lf*lf
      //
      + (-8./5.*ZETA2*ZETA2*CA*CA*CA-2*beta0*gB1-2*d1*ZETA2*CA*CA+2*beta0*beta1+beta0*beta0*d1-4*ZETA2*CA*gA1-12*beta0*CA*CA*ZETA3)*lf*lf
      //+ (-4.*CA*gA1*ZETA2-2.*beta0*(gB1-beta1+6.*CA*CA*ZETA3)+beta0*beta0*(2.*CA*ZETA2+W1)-2.*CA*CA*W1*ZETA2+4.*CA*CA*CA*(ZETA2*ZETA2-6.*ZETA4))*lf*lf
      //
      + (134./9.*CA*CA*CA*ZETA3-39*ZETA2*CA*CA*CA*ZETA3-3*d1*beta1-28./27.*nf*ZETA2*CA*CA+2*d1*gB1+96*ZETA5*CA*CA*CA
      +202./27.*ZETA2*CA*CA*CA-36*ZETA2*CA*CA*CA*ZETA3+8*d1*CA*CA*ZETA3-20./9.*nf*CA*CA*ZETA3+32./5.*beta0*ZETA2*ZETA2*CA*CA
      -22./3.*ZETA2*ZETA2*CA*CA*CA-2*beta0*d2-2*beta2+8*CA*ZETA3*gA1+4./3.*nf*ZETA2*ZETA2*CA*CA+2*gB2)*lf;
      d2 += -2.*CA*CA*ZETA2*lf*lf + (-3./4. + 171./2.*ZETA3 + 1./4.*nf - beta0*d1)*lf;
      //
      // adding renormalization scale dependence
      d3 += (2*beta2+3*beta1*d1+4*beta0*d2)*lrf + (7*beta0*beta1+6*beta0*beta0*d1)*lrf*lrf + 4*Power(beta0*lrf,3);
      d2 += (2*beta1+3*beta0*d1)*lrf + 3*beta0*beta0*lrf*lrf;
      d1 += 2*beta0*lrf;
      //
      d1 = PLd1 + W1;
      d2 = PLd2 + PLd1*W1 + W2;
      d3 = PLd3 + PLd2*W1 + PLd1*W2 + W3;
      */
      //
      //double a11 = 4*CA;
      double a23 = 8*CA*CA;
      double a22 = -4*CA*beta0;
      double a21 = 4*CA*d1 + CA*CA*(67./9.-18.*ZETA2) - 10./9.*CA*nf;
      double a20 = CA*CA*(-101./27.+11./3.*ZETA2+39./2.*ZETA3) + CA*nf*(14./27.-2./3.*ZETA2);
      // factorization scale dependence (as=as(M))
      d3 += (4*beta0*beta0*beta0 - 6*CA*CA*beta0*ZETA2 + 8./3.*CA*CA*CA*ZETA3)*lf*lf*lf
	//
	//+ (-2*CA*CA*ZETA2*d1-8./5.*CA*CA*CA*ZETA2*ZETA2+3*beta0*beta0*d1+beta0*beta1+20*CA*CA*ZETA3*beta0+6*beta0*gB1-4*CA*ZETA2*gA1) * lf*lf
	+ (beta0*(6*gB1+beta1) - 4*CA*gA1*ZETA2 + 20*CA*CA*beta0*ZETA3 + 8*CA*CA*CA*(ZETA2*ZETA2-3*ZETA4) + d1*(3*beta0*beta0-2*CA*CA*ZETA2))*lf*lf
	//
	//+ (32./5.*CA*CA*beta0*ZETA2*ZETA2+2*beta0*d2+2*gB1*d1-28./27.*CA*CA*nf*ZETA2+202./27.*CA*CA*CA*ZETA2+8*CA*CA*ZETA3*d1+4./3.*CA*CA*nf*ZETA2*ZETA2-22./3.*CA*CA*CA*ZETA2*ZETA2+96*CA*CA*CA*ZETA5-20./9.*CA*CA*ZETA3*nf+134./9.*CA*CA*CA*ZETA3-36*CA*CA*CA*ZETA3*ZETA2+8*CA*ZETA3*gA1+2*gB2-39*CA*CA*CA*ZETA3*ZETA2) *lf;
	+ (2*gB2+8*CA*gA1*ZETA3+2*gB1*d1+2*beta0*d2+2*CA*(6*ZETA5*a23-2*ZETA4*a22+ZETA3*a21-ZETA2*a20))*lf;
      d2 += (-2.*CA*CA*ZETA2 + 3*beta0*beta0)*lf*lf + (2*gB1+2*beta0*d1+8*CA*CA*ZETA3)*lf;
      d1 += 2*beta0*lf;
      //
      // adding renormalization scale dependence
      d3 += (2*beta2+3*beta1*d1+4*beta0*d2)*lr + (7*beta0*beta1+6*beta0*beta0*d1)*lr*lr + 4*Power(beta0*lr,3);
      d2 += (2*beta1+3*beta0*d1)*lr + 3*beta0*beta0*lr*lr;
      d1 += 2*beta0*lr;
      //
      // restore normalization
      d1 /= M_PI;
      d2 /= M_PI*M_PI;
      d3 /= M_PI*M_PI*M_PI;
      //
      // construct full g0k
      //const double c1 = EulerGamma;
      double c2 = Power(EulerGamma,2) -ZETA2;
      double c3 = Power(EulerGamma,3) -3*EulerGamma*ZETA2 +2*ZETA3;
      double c4 = Power(EulerGamma,4) -6*EulerGamma*EulerGamma*ZETA2 +8*EulerGamma*ZETA3 +3*ZETA2*ZETA2 -6*ZETA4;
      double c5 = 24*ZETA5 +3*EulerGamma*ZETA2*ZETA2 +Power(EulerGamma,5) - 10*Power(EulerGamma,3)*ZETA2 +20*EulerGamma*EulerGamma*ZETA3 -20*ZETA3*ZETA2;
      double c6 = 144*ZETA5*EulerGamma +Power(EulerGamma,6) -15*Power(EulerGamma,4)*ZETA2 +40*Power(EulerGamma,3)*ZETA3 +9*EulerGamma*EulerGamma*ZETA2*ZETA2 -120*EulerGamma*ZETA3*ZETA2 +40*ZETA3*ZETA3 -45./7.*Power(ZETA2,3);
      //
      double S1 = 2*A1* (ZETA2+EulerGamma*EulerGamma - EulerGamma*lf);
      double S2 = 4*A1*b0/3.*c3 -2*( A1*b0*(2*EulerGamma+lr) + A2 )*c2
	- EulerGamma* ( D2 - 4.*A2*EulerGamma - 4.*A1*b0*(EulerGamma*EulerGamma+ZETA2)
			- 4.*A1*b0*EulerGamma*lr +2*A2*lf
			- A1*b0*(lr*lr - lrf*lrf) );
      double S3 = -c4* (4*A1*Power(b0,2))/3.
	+c3* (4*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma) - 4*A1*b0*log(M/muR)))/3.
	-c2* 2*(A3 + b0*(-D2 + 4*A2*EulerGamma + 
			 2*A1*(b1*EulerGamma + 2*b0*(Power(EulerGamma,2) + ZETA2))) - 
		2*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma))*log(M/muR) + 
		4*A1*Power(b0,2)*Power(log(M/muR),2))
	-EulerGamma* (3*D3 - 12*A3*EulerGamma + 12*b0*D2*EulerGamma - 24*A2*b0*Power(EulerGamma,2) - 
		      12*A1*b0*b1*Power(EulerGamma,2) - 16*A1*Power(b0,2)*Power(EulerGamma,3) - 
		      24*A2*b0*ZETA2 - 12*A1*b0*b1*ZETA2 - 48*A1*Power(b0,2)*EulerGamma*ZETA2 - 
		      32*A1*Power(b0,2)*ZETA3 + 12*
		      (A3 + b0*(-D2 + 4*A2*EulerGamma + 
				2*A1*(b1*EulerGamma + 2*b0*(Power(EulerGamma,2) + ZETA2))))*log(M/muR) - 
		      12*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma))*Power(log(M/muR),2) + 
		      16*A1*Power(b0,2)*Power(log(M/muR),3) - 12*A3*log(muF/muR) + 
		      12*b0*(2*A2 + A1*b1)*Power(log(muF/muR),2) - 
		      16*A1*Power(b0,2)*Power(log(muF/muR),3))/3.;
      // S1^2
      const double S1s = -4*A1*A1*(c4 -2*(2*EulerGamma-lf)*c3 +Power(2*EulerGamma-lf,2)*c2);
      // S1^3
      const double S1c = -8*A1*A1*A1*(c6 -3*(2*EulerGamma-lf)*c5 +3*Power(2*EulerGamma-lf,2)*c4 -Power(2*EulerGamma-lf,3)*c3);
      // S1*S2
      const double S1S2 = 2*A1*(4*A1*b0/3.*c5 -2*( A1*b0*(2*EulerGamma+lr) + A2 )*c4
				- c3* ( D2 - 4.*A2*EulerGamma - 4.*A1*b0*(EulerGamma*EulerGamma+ZETA2)
					- 4.*A1*b0*EulerGamma*lr +2*A2*lf
					- A1*b0*(lr*lr - lrf*lrf) ))
	-2*A1*(2*EulerGamma-lf)* (4*A1*b0/3.*c4 -2*( A1*b0*(2*EulerGamma+lr) + A2 )*c3
				  - c2* ( D2 - 4.*A2*EulerGamma - 4.*A1*b0*(EulerGamma*EulerGamma+ZETA2)
					  - 4.*A1*b0*EulerGamma*lr +2*A2*lf
					  - A1*b0*(lr*lr - lrf*lrf) ));
      //
      const double gD1 = S1;
      const double gD2 = S1s/2. + S2 + gD1*gD1;
      const double gD3 = S1c/6. + S1S2 + S3 + 2*gD1*gD2 - gD1*gD1*gD1;
      //
      g01 = d1 + gD1;
      g02 = d2 + d1*gD1 + gD2;
      g03 = d3 + d2*gD1 + d1*gD2 + gD3; 
      //
      g0sub1s = S1;
      g0sub2s = g01*S1 + S1s/2. + S2;
      g0sub3s = g01*(S1s/2.+S2) + g02*S1 + S1c/6. + S1S2 + S3;
      //
      /*/
      c2 += ZETA2;
      c3 += -(1.5)*ZETA3;
      c4 += 27./5.*ZETA2*ZETA2;
      c5 += -(22.5)*ZETA5-(15.0)*ZETA2*ZETA3;
      c6 += 549./7.*ZETA2*ZETA2*ZETA2+(22.5)*ZETA3*ZETA3;
      //
      S1 = 2*A1* (ZETA2+EulerGamma*EulerGamma - EulerGamma*lf);
      S2 = 4*A1*b0/3.*c3 -2*( A1*b0*(2*EulerGamma+lr) + A2 )*c2
	- EulerGamma* ( D2 - 4.*A2*EulerGamma - 4.*A1*b0*(EulerGamma*EulerGamma+ZETA2)
			- 4.*A1*b0*EulerGamma*lr +2*A2*lf
			- A1*b0*(lr*lr - lrf*lrf) );
      S3 = -c4* (4*A1*Power(b0,2))/3.
	+c3* (4*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma) - 4*A1*b0*log(M/muR)))/3.
	-c2* 2*(A3 + b0*(-D2 + 4*A2*EulerGamma + 
			 2*A1*(b1*EulerGamma + 2*b0*(Power(EulerGamma,2) + ZETA2))) - 
		2*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma))*log(M/muR) + 
		4*A1*Power(b0,2)*Power(log(M/muR),2))
	-EulerGamma* (3*D3 - 12*A3*EulerGamma + 12*b0*D2*EulerGamma - 24*A2*b0*Power(EulerGamma,2) - 
		      12*A1*b0*b1*Power(EulerGamma,2) - 16*A1*Power(b0,2)*Power(EulerGamma,3) - 
		      24*A2*b0*ZETA2 - 12*A1*b0*b1*ZETA2 - 48*A1*Power(b0,2)*EulerGamma*ZETA2 - 
		      32*A1*Power(b0,2)*ZETA3 + 12*
		      (A3 + b0*(-D2 + 4*A2*EulerGamma + 
				2*A1*(b1*EulerGamma + 2*b0*(Power(EulerGamma,2) + ZETA2))))*log(M/muR) - 
		      12*b0*(2*A2 + A1*(b1 + 4*b0*EulerGamma))*Power(log(M/muR),2) + 
		      16*A1*Power(b0,2)*Power(log(M/muR),3) - 12*A3*log(muF/muR) + 
		      12*b0*(2*A2 + A1*b1)*Power(log(muF/muR),2) - 
		      16*A1*Power(b0,2)*Power(log(muF/muR),3))/3.;
      //*/
      g0bar1 = d1;
      g0bar2 = g02 - g01*S1 - S2 + S1*S1/2.;
      g0bar3 = g03 - g02*S1 - g01*(S2-S1*S1/2.) - (S1*S1*S1/6. - S1*S2 + S3);
    }
    //
    //cout << "gD1 = " << gD1 << endl;
    //cout << "gD2 = " << gD2 << endl;
    //cout << "gD3 = " << gD3 << endl;
    //cout << "g01 = " << g01 << endl;
    //cout << "g02 = " << g02 << endl;
    //cout << "g03 = " << g03 << endl;
    //
    //
    g0sub1 = 0;
    g0sub2 = 0;
    g0sub3 = 0;
    //
    if(!expBP && expDconst && prescr>0) {
      g0sub1 = g0sub1s;
      g0sub2 = g0sub2s;
      g0sub3 = g0sub3s;
    }

    if(expBP && expDconst) {
      //cout << "g0bar1 = " << g0bar1 << endl;
      //cout << "g0bar2 = " << g0bar2 << endl;
      //cout << "g0bar3 = " << g0bar3 << endl;
      g01 = g0bar1;
      g02 = g0bar2;
      g03 = g0bar3;
    }





    /*
    //  old version (valid only within the heavy top EFT)
    g01 = CA*(4.*ZETA2+2.*EulerGamma*EulerGamma + 4.*EulerGamma*log(muF/M) + 4./3.*M_PI*b0*log(muR/muF))/M_PI + 11./2/M_PI;
    //
    double delta_g02 = 11399./144. + 133./2.*ZETA2 - 9./20.*ZETA2*ZETA2 - 165./4.*ZETA3
      + (19./8.+2.*nf/3.)*2.*log(M/mtop) + nf*(-1189./144.-5./3.*ZETA2+5./6.*ZETA3)
      + 3.*b0*b0*M_PI*M_PI *pow(2.*log(muF/muR),2) - 18.*ZETA2*pow(2.*log(M/muF),2)
      + (169./4. + 171./2.*ZETA3 - 19./6.*nf + 12.*M_PI*b0*ZETA2) * 2.*log(M/muF)
      + (-465./8. + 13./3.*nf - 18.*M_PI*b0*ZETA2) * 2.*log(M/muR);
    //
    g02 = delta_g02 + EulerGamma*(101./3.-14./9.*nf-63./2.*ZETA3)
      + EulerGamma*EulerGamma*(133./2.-5./3.*nf+63.*ZETA2)
      + EulerGamma*EulerGamma*EulerGamma*4.*M_PI*b0
      + 18.* pow(EulerGamma,4)
      + (133./2.-5./3.*nf)*ZETA2
      + 261./2.*ZETA4 + (22.-4./3.*nf)*ZETA3
      + ((-165./4.+5./2.*nf)*EulerGamma+18.*EulerGamma*EulerGamma+18.*ZETA2) * pow(2.*log(M/muF),2)
      + 18.*M_PI*EulerGamma*b0 * 4.*log(M/muF)*log(M/muR)
      - 18.*M_PI*b0 * (EulerGamma*EulerGamma+ZETA2) * 2.*log(M/muR) /// pow(2.*log(M/muR),2)  ////// !!!! no power 2!!
      + (12.*(M_PI*b0-3.*EulerGamma)*EulerGamma*EulerGamma + (-133./2.+5./3.*nf-63.*ZETA2)*EulerGamma
	 + 12.*M_PI*b0*ZETA2 - 72.*ZETA3) * 2.*log(M/muF);
    g02 /= M_PI*M_PI;
    */

    //
    return;
  }


};






